import 'dart:async';
import 'dart:convert';
import 'package:flutter/cupertino.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_miniprogram_core/nui_miniprogram_enum.dart';
import 'package:nui_miniprogram_core/nui_miniprogram_message.dart';

//Concept
/*
1) Send a message
2) Wait for response
3)
 */

typedef NUICrossEventMessageListener = Future<NUIMiniProgramMessage> Function(NUIMiniProgramMessage message);
typedef NUIMiniProgramChannelListener = Future<NUIMiniProgramMessage> Function(String data, NUIMiniProgramMessage message);

class NUIMiniProgramChannel{
  NUIMiniProgramChannel(String channel, {NUIMiniProgramChannelListener listener}){
    this._channel = channel;
    this._listener = listener;
  }
  String _channel;
  NUIMiniProgramChannelListener _listener;

  String get channel => _channel;
}


abstract class NUIMiniProgramJSEngineInterface{
  Map<String, Completer<NUIMiniProgramMessage>> _messageQueue = Map();
  List<NUIMiniProgramChannel> _channels = [];
  List<NUIMiniProgramChannel> _internalChannels = [];

  void initInternalChannelsForWeb(BuildContext context) async{
    _internalChannels = [
      NUIMiniProgramChannel(NUIMiniProgramWebEventType.CAN_POP.value, listener: (data, message) async{
        final canPop = Navigator.canPop(context);
        if(canPop){
          pop(context);
        }
        final reply = NUIMiniProgramMessage(id: message.id, data: canPop ? "true": "false", channel: NUIMiniProgramWebEventType.CAN_POP.value);
        return reply;
      }),
    ];
  }

  void setChannels(List<NUIMiniProgramChannel> channels){
    this._channels.clear();
    this._channels.addAll(channels);
  }

  List<NUIMiniProgramChannel> get channels => _channels;

  void messageReceived(String messageRaw) async{
    final dataMessage = processEventMessage(messageRaw);
    if(dataMessage != null) {
      completeMessageInQueue(dataMessage);
      final reply = await _redirectPlatformMessage(dataMessage);
      if(reply != null){
        await _sendCrossEvent(reply);
      }
    }
  }

  Future<NUIMiniProgramMessage> _redirectPlatformMessage(NUIMiniProgramMessage message) async{
    //The message is a received response, so ignore this as an internal message
    if(message.isResponseReceived()){
      print("NUIMiniProgram: Just a response received message, so ignore and stop the conversation");
      return null;
    }

    for(var element in _internalChannels){
      if(match(element.channel, message.channel)){
        if (element._listener != null) {
          //User has a custom listener to this event, check if there is a custom response or use the default received response
          final result = element._listener(message.data, message);
          return result ?? message.responseReceived();
        }
        else{
          return message.responseReceived();
        }
      }
    }

    for(var element in _channels){
      if(match(element.channel, message.channel)) {
        if (element._listener != null) {
          //User has a custom listener to this event, check if there is a custom response or use the default received response
          final result = element._listener(message.data, message);
          return result ?? message.responseReceived();
        }
        else{
          return message.responseReceived();
        }
      }
    }
    return message.responseReceived();
  }

  void addMessageToQueue(NUIMiniProgramMessage message){
    _messageQueue[message.id] = Completer();
  }

  void completeMessageInQueue(NUIMiniProgramMessage message){
    final completed = _messageQueue[message.id];
    if(completed != null && completed.isCompleted){
      _messageQueue[message.id] = Completer();
    }
    _messageQueue[message.id]?.complete(message);
  }

  Completer<NUIMiniProgramMessage> _getResponseMessageQueue(String id){
    final messageQueue = _messageQueue[id];
    if(messageQueue == null) return null;
    return messageQueue;
  }

  Future<bool> sendPlatformCrossEvent(NUIMiniProgramMessage message);

  Future<bool> _sendCrossEvent(NUIMiniProgramMessage message) async{
    try {
      addMessageToQueue(message);
      await sendPlatformCrossEvent(message);
      return true;
    }catch(e){
      print("[NUIMiniProgramJSEngineInterface] Failed to send cross event with error : $e");
      return false;
    }
  }

  Future<NUIMiniProgramMessage> sendAndAwaitCrossEvent(NUIMiniProgramMessage message) async{
    final sent = await _sendCrossEvent(message);
    if(sent) {
      final queue = _getResponseMessageQueue(message.id);
      if (queue == null) return null;
      return queue.future;
    }
    return null;
  }

  NUIMiniProgramMessage processEventMessage(String message){
    try{
      final decoded = jsonDecode(message);
      final programMessage = NUIMiniProgramMessage().fromJson(decoded);
      if(programMessage.channel == null || programMessage.data == null){
        return null;
      }
      return programMessage;
    }catch(e){
      return null;
    }
  }
}